import glob
import os
import re
import zipfile
from abc import ABC
from datetime import datetime, timezone

import numpy as np
import mne
from moabb.datasets import download as dl
from moabb.datasets.base import BaseDataset

VSPELL_BASE_URL = "https://zenodo.org/record/"
VISUAL_SPELLER_LLP_URL = VSPELL_BASE_URL + "5831826/files/"
VISUAL_SPELLER_MIX_URL = VSPELL_BASE_URL + "5831879/files/"
OPTICAL_MARKER_CODE = 500
VALID_LETTER_NUMBERS = []
VALID_LETTER_NUMBERS.extend([i for i in range(201, 242 + 1)])
VALID_LETTER_NUMBERS.extend([i for i in range(101, 142 + 1)])

NUMBER_TO_LETTER_DICT = {
    1: "A",
    2: "B",
    3: "#",
    4: "C",
    5: "#",
    6: "D",
    7: "E",
    8: "#",
    9: "F",
    10: "G",
    11: "H",
    12: "I",
    13: "J",
    14: "#",
    15: "K",
    16: "L",
    17: "M",
    18: "#",
    19: "N",
    20: "O",
    21: "P",
    22: "Q",
    23: "#",
    24: "R",
    25: "S",
    26: "T",
    27: "U",
    28: "#",
    29: "V",
    30: "W",
    31: "X",
    32: "Y",
    33: "#",
    34: "Z",
    35: " ",
    36: "#",
    37: ".",
    38: "#",
    39: ",",
    40: "!",
    41: "?",
    42: "<",
}


def seq_labels_from_epoch(epo):
    s1 = epo["Sequence_1"].events
    s1[:, 2] = 1
    s2 = epo["Sequence_2"].events
    s2[:, 2] = 2
    s = np.vstack([s1, s2])
    l1 = epo["Target"].events
    l1[:, 2] = 1
    l0 = epo["NonTarget"].events
    l0[:, 2] = 0
    l = np.vstack([l0, l1])
    s = np.array(sorted(s, key=lambda x: x[0]))
    l = np.array(sorted(l, key=lambda x: x[0]))
    return s[:, 2], l[:, 2]


def natural_sort(l):
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(l, key=alphanum_key)


class _BaseVisualMatrixSpellerDataset(BaseDataset, ABC):
    def __init__(
        self,
        n_subjects,
        run_split_n,
        raw_slice_offset,
        data_path=None,
        src_url=None,
        **kwargs,
    ):
        if run_split_n is not None and (run_split_n > 7 or run_split_n < 1):
            raise ValueError(f"Split must be in interval [1, 7] or None. But was {run_split_n}.")

        self.n_channels = 31  # all channels except 5 times x_* CH and EOGvu
        if kwargs["interval"] is None:
            # "Epochs were windowed to [−200, 700] ms relative to the stimulus onset [...]."
            kwargs["interval"] = [-0.2, 0.7]

        super().__init__(
            events=dict(Target=10002, NonTarget=10001),
            paradigm="p300",
            subjects=(np.arange(n_subjects) + 1).tolist(),
            **kwargs,
        )

        self.run_split_n = run_split_n
        self.raw_slice_offset = 2_000 if raw_slice_offset is None else raw_slice_offset
        self._src_url = src_url
        self._data_path = data_path

    @staticmethod
    def _filename_trial_info_extraction(vhdr_file_path):
        vhdr_file_name = os.path.basename(vhdr_file_path)
        run_file_pattern = "^matrixSpeller_Block([0-9]+)_Run([0-9]+)\\.vhdr$"
        vhdr_file_patter_match = re.match(run_file_pattern, vhdr_file_name)

        if not vhdr_file_patter_match:
            # TODO: raise a wild exception?
            print(vhdr_file_path)

        session_name = os.path.basename(os.path.dirname(vhdr_file_path))
        block_idx = vhdr_file_patter_match.group(1)
        run_idx = vhdr_file_patter_match.group(2)
        return session_name, block_idx, run_idx

    def _get_single_subject_data(self, subject):
        subject_data_vhdr_files = self.data_path(subject)
        sessions = dict()

        for file_idx, subject_data_vhdr_file in enumerate(subject_data_vhdr_files):
            (
                session_name,
                block_idx,
                run_idx,
            ) = VisualMatrixSpellerLLPDataset._filename_trial_info_extraction(
                subject_data_vhdr_file
            )
            raw_bvr_list = _read_raw_llp_study_data(
                vhdr_fname=subject_data_vhdr_file,
                run_split_n=self.run_split_n,
                raw_slice_offset=self.raw_slice_offset,
                run_idx=run_idx,
                verbose=None,
            )

            session_name = f"{session_name}_block_{block_idx}"
            if self.run_split_n is not None:
                for split_idx, raw_split in enumerate(raw_bvr_list):
                    session_name = session_name + f"_run_{run_idx}_split_{split_idx}"
                    sessions[session_name] = {"0": raw_split}
            else:
                if session_name not in sessions.keys():
                    sessions[session_name] = dict()
                sessions[session_name][run_idx] = raw_bvr_list[0]

        return sessions

    def data_path(self, subject, path=None, force_update=False, update_path=None, verbose=None):

        url = f"{self._src_url}subject{subject:02d}.zip"
        data_archive_path = dl.data_path(url, "llp")
        data_dir_extracted_path = os.path.dirname(data_archive_path)
        # else:
        #     raise ValueError(f'URL or data path must be given but both are None.')

        subject_dir_path = os.path.join(data_dir_extracted_path, f"subject{subject:02d}")

        data_extracted = os.path.isdir(subject_dir_path)
        if not data_extracted:
            # print('unzip', path_to_data_archive)  # TODO logging? check verbose
            zipfile_path = glob.glob(
                os.path.join(data_dir_extracted_path, data_archive_path, "*.zip")
            )[0]
            _BaseVisualMatrixSpellerDataset._extract_data(data_dir_extracted_path, zipfile_path)

        run_glob_pattern = os.path.join(
            data_dir_extracted_path,
            f"subject{subject:02d}",
            "matrixSpeller_Block*_Run*.vhdr",
        )
        subject_paths = glob.glob(run_glob_pattern)

        return natural_sort(subject_paths)

    @staticmethod
    def _extract_data(data_dir_extracted_path, data_archive_path):
        zip_ref = zipfile.ZipFile(data_archive_path, "r")
        zip_ref.extractall(data_dir_extracted_path)

    def load_epochs(self, block_nrs=(1, 2, 3), fband=(0.5, 16), sampling_rate=100):
        raws = list()
        event_list = list()
        block_ids = list()
        for subj_id, subj_data in self.get_data().items():
            for session_id, session_data in subj_data.items():
                for run_id, run_data in session_data.items():
                    bnr = int(run_data.filenames[0].split("Block")[1][0])
                    if bnr in block_nrs:
                        raws.append(run_data)
                        events, _ = mne.events_from_annotations(run_data)
                        event_list.append(events)
                        block_ids.extend([bnr] * len(events))

        DUMMY_DATESTR = "20100101120150667372"
        meas_date = datetime.strptime(DUMMY_DATESTR, "%Y%m%d%H%M%S%f")
        meas_date = meas_date.replace(tzinfo=timezone.utc)
        [r.set_meas_date(meas_date) for r in raws]
        raws_epochs, _ = mne.concatenate_raws(raws, events_list=event_list, preload=True)
        raws_epochs = raws_epochs.filter(*fband)
        events, events_dict = mne.events_from_annotations(raws_epochs)

        epochs = mne.Epochs(
            raw=raws_epochs,
            events=events,
            event_id=events_dict,
            picks=["eeg"],
            preload=True,
            tmin=self.tmin,
            tmax=self.tmax,
            baseline=None,
        )
        epochs = epochs.resample(sampling_rate)
        # symbols = pd.unique(list(NUMBER_TO_LETTER_DICT.values()))
        # cols = ["block_nr", "letter_nr", *symbols]
        # md = pd.DataFrame(columns=cols)
        # for i in range(len(epochs)):
        #     print(str(i))
        #     e = epochs[i]
        #     ev = list(e.event_id.keys())[0].split("/")
        #     row = pd.Series(index=cols)
        #     row[:] = False
        #     highlights = [e for e in ev if len(e) == 1]
        #     row.loc[highlights] = True
        #     row.block_nr = block_ids[i]
        #     row.letter_nr = int(ev[0].split("_")[-1])
        #     md = md.append(row, ignore_index=True)
        #
        # epochs.metadata = md

        return epochs, raws_epochs


class VisualMatrixSpellerLLPDataset(_BaseVisualMatrixSpellerDataset):
    """
    Learning from label proportions for a visual matrix speller (ERP) dataset from Hübner et al 2017 [1]_.


    **Dataset description**

    The subjects were asked to spell the sentence: “Franzy jagt im komplett verwahrlosten Taxi quer durch Freiburg”.
    The sentence was chosen because it contains each letter used in German at least once. Each subject spelled this
    sentence three times. The stimulus onset asynchrony (SOA) was 250 ms (corresponding to 15 frames on the LCD screen
    utilized) while the stimulus duration was 100 ms (corresponding to 6 frames on the LCD screen utilized). For each
    character, 68 highlighting events occurred and a total of 63 characters were spelled three times. This resulted in
    a total of 68 ⋅ 63 ⋅ 3 = 12852 EEG epochs per subject. Spelling one character took around 25 s including 4 s for
    cueing the current symbol, 17 s for highlighting and 4 s to provide feedback to the user. Assuming a perfect
    decoding, these timing constraints would allow for a maximum spelling speed of 2.4 characters per minute. Fig 1
    shows the complete experimental structure and how LLP is used to reconstruct average target and non-target ERP
    responses.

    Subjects were placed in a chair at 80 cm distance from a 24-inch flat screen. EEG signals from 31 passive Ag/AgCl
    electrodes (EasyCap) were recorded, which were placed approximately equidistantly according to the extended
    10–20 system, and whose impedances were kept below 20 kΩ. All channels were referenced against the nose and the
    ground was at FCz. The signals were registered by multichannel EEG amplifiers (BrainAmp DC, Brain Products) at a
    sampling rate of 1 kHz. To control for vertical ocular movements and eye blinks, we recorded with an EOG electrode
    placed below the right eye and referenced against the EEG channel Fp2 above the eye. In addition, pulse and
    breathing activity were recorded.

    Parameters
    ----------
    run_split_n: None, int
        Define how to split the runs. If None load whole EEG sessions. If integer load splits of run_split_n runs.
        That is to say runs are loaded individually if run_split_n is set to one. If run_split_n is set to 2 then the
        data is split into two splits of three trials a 68 epochs. The seventh trial is discarded. Default is set to
        None.
        Note: the data can be split into at most seven splits.
    interval: array_like
        range/interval in milliseconds in which the brain response/activity relative to an event/stimulus onset lies in.
        Default is set to [-.2, .7].
    raw_slice_offset: int, None
        defines the crop offset in milliseconds before the first and after the last event (target or non-targeet) onset.
        Default None which crops with an offset 2,000 ms.

    References
    ----------
    .. [1] Hübner, D., Verhoeven, T., Schmid, K., Müller, K. R., Tangermann, M., & Kindermans, P. J. (2017)
           Learning from label proportions in brain-computer interfaces: Online unsupervised learning with guarantees.
           PLOS ONE 12(4): e0175856.
           https://doi.org/10.1371/journal.pone.0175856
    """

    def __init__(
        self,
        sessions_per_subject=1,
        data_path=None,
        run_split_n=None,
        interval=None,
        raw_slice_offset=None,
    ):
        llp_speller_paper_doi = "10.1371/journal.pone.0175856"
        self.tmin = -0.2
        self.tmax = 0.8
        self.downsample_freq = 100
        super().__init__(
            src_url=VISUAL_SPELLER_LLP_URL,
            data_path=data_path,
            run_split_n=run_split_n,
            raw_slice_offset=raw_slice_offset,
            n_subjects=13,
            sessions_per_subject=sessions_per_subject,  # if varying, take minimum
            code="Visual Speller LLP",
            interval=interval,
            doi=llp_speller_paper_doi,
        )


class VisualMatrixSpellerMixDataset(_BaseVisualMatrixSpellerDataset):
    """
    Mixtue of LLP and EM for a visual matrix speller (ERP) dataset from Hübner et al 2018 [1]_.

    Within a single session, a subject was asked to spell the beginning of a sentence in each of three blocks.The text
    consists of the 35 symbols “Franzy jagt im Taxi quer durch das”. Each block, one of the three decoding
    algorithms (EM, LLP, MIX) was used in order to guess the attended symbol. The order of the blocks was
    pseudo-randomized over subjects, such that each possible order of the three decoding algorithms was used twice.
    This randomization should reduce systematic biases by order effects or temporal effects, e.g., due to fatigue or
    task-learning.

    A trial describes the process of spelling one character. Each of the 35 trials per block contained 68 highlighting
    events. The stimulus onset asynchrony (SOA) was 250 ms and the stimulus duration was 100 ms leading to an
    interstimulus interval (ISI) of 150 ms.

    Parameters
    ----------
    run_split_n: None, int
        Define how to split the runs. If None load whole EEG sessions. If integer load splits of run_split_n runs.
        That is to say runs are loaded individually if run_split_n is set to one. If run_split_n is set to 2 then the
        data is split into two splits of three trials a 68 epochs. The seventh trial is discarded. Default is set to
        None.
        Note: the data can be split into at most seven splits.
    interval: array_like
        range/interval in milliseconds in which the brain response/activity relative to an event/stimulus onset lies in.
        Default is set to [-.2, .7].
    raw_slice_offset: int, None
        defines the crop offset in milliseconds before the first and after the last event (target or non-targeet) onset.
        Default None which crops with an offset 2,000 ms.

    References
    ----------
    .. [1] Huebner, D., Verhoeven, T., Mueller, K. R., Kindermans, P. J., & Tangermann, M. (2018).
           Unsupervised learning for brain-computer interfaces based on event-related potentials: Review and online
           comparison [research frontier].
           IEEE Computational Intelligence Magazine, 13(2), 66-77.
           https://doi.org/10.1109/MCI.2018.2807039
    """

    def __init__(self, run_split_n=None, interval=None, raw_slice_offset=None):
        mix_speller_paper_doi = "10.1109/MCI.2018.2807039"
        self.tmin = -0.2
        self.tmax = 0.8
        self.downsample_freq = 100
        super().__init__(
            src_url=VISUAL_SPELLER_MIX_URL,
            run_split_n=run_split_n,
            raw_slice_offset=raw_slice_offset,
            n_subjects=12,
            sessions_per_subject=1,  # if varying, take minimum
            code="Visual Speller MIX",
            interval=interval,
            doi=mix_speller_paper_doi,
        )


def _read_raw_llp_study_data(vhdr_fname, run_split_n, raw_slice_offset, run_idx, verbose=None):
    """
    Read LLP BVR recordings file. Ignore the different sequence lengths. Just tag event as target or non-target if it
    contains a target or does not contain a target.

    Parameters
    ----------
    vhdr_fname: str
        Path to the EEG header file.
    n_chars: number of characters of the spelled sentence
    verbose : bool, int, None
        specify the loglevel.

    Returns
    -------
    raw_object: mne.io.Raw
        the loaded BVR raw object.
    """
    non_scalp_channels = ["EOGvu", "x_EMGl", "x_GSR", "x_Respi", "x_Pulse", "x_Optic"]
    raw_bvr = mne.io.read_raw_brainvision(
        vhdr_fname=vhdr_fname,  # eog='EOGvu',
        misc=non_scalp_channels,
        preload=True,
        verbose=verbose,
    )  # type: mne.io.Raw
    raw_bvr.set_montage("standard_1020")
    # TODO: MNE/MOABB do not like anonymized data, therefore we need to set a dummy date
    # DUMMY_DATESTR = '20100101120150667372'
    # meas_date = datetime.strptime(DUMMY_DATESTR, '%Y%m%d%H%M%S%f')
    # meas_date = meas_date.replace(tzinfo=timezone.utc)
    # raw_bvr.set_meas_date(meas_date)

    events = _parse_events(raw_bvr)

    (
        onset_arr_list,
        marker_arr_list,
        sequence_arr_list,
        letter_arr_list,
    ) = _extract_target_non_target_description(events, run_split_n)

    has_multiple_splits = run_split_n is not None and run_split_n > 1

    def annotate_and_crop_raw(onset_arr, marker_arr, sequence_arr):
        raw = raw_bvr.copy() if has_multiple_splits else raw_bvr
        raw_annotated = raw.set_annotations(
            _create_annotations_from(
                marker_arr, onset_arr, sequence_arr, letter_arr_list, raw, run_idx
            )
        )

        tmin = max((onset_arr[0] - raw_slice_offset) / 1e3, 0)
        tmax = min((onset_arr[-1] + raw_slice_offset) / 1e3, raw.times[-1])
        return raw_annotated.crop(tmin=tmin, tmax=tmax, include_tmax=True)

    return list(map(annotate_and_crop_raw, onset_arr_list, marker_arr_list, sequence_arr_list))


def _create_annotations_from(marker_arr, onset_arr, sequence_arr, letter_arr, raw_bvr, run_idx):
    default_bvr_marker_duration = raw_bvr.annotations[0]["duration"]

    onset = onset_arr / 1e3  # convert onset in seconds to ms
    durations = np.repeat(default_bvr_marker_duration, len(marker_arr))
    description = _create_description(marker_arr, sequence_arr, letter_arr, run_idx)

    orig_time = raw_bvr.annotations[0]["orig_time"]
    return mne.Annotations(
        onset=onset, duration=durations, description=description, orig_time=orig_time
    )


def _create_description(marker_arr, sequence_arr, letter_arr, run_idx):
    desc = list()
    epochs_per_trial = 68
    trials_per_run = 7
    letter_nr = (trials_per_run * (int(run_idx) - 1)) + 1
    for i in range(0, trials_per_run):
        start_idx = i * epochs_per_trial
        end_idx = (i + 1) * epochs_per_trial

        for label, seq, letters in zip(
            marker_arr[start_idx:end_idx],
            sequence_arr[start_idx:end_idx],
            letter_arr[0][start_idx:end_idx],
        ):
            single_desc = f"Letter_{letter_nr}/"
            if seq == 21:
                single_desc += "Sequence_1/"
            elif seq == 22:
                single_desc += "Sequence_2/"

            if label == 0:
                single_desc += "NonTarget/"
            elif label == 1:
                single_desc += "Target/"
            for letter in letters:
                letter = int(str(letter)[-2:])
                if letter != -1:
                    single_desc += NUMBER_TO_LETTER_DICT[letter] + "/"
            desc.append(single_desc[:-1])
        letter_nr += 1
    return desc


def _parse_events(raw_bvr):
    stimulus_pattern = re.compile("(Stimulus/S|Optic/O) *([0-9]+)")

    def parse_marker(desc):
        match = stimulus_pattern.match(desc)
        if match is None:
            return None
        if match.group(1) == "Optic/O":
            return OPTICAL_MARKER_CODE
        return int(match.group(2))

    events, _ = mne.events_from_annotations(raw=raw_bvr, event_id=parse_marker, verbose=None)
    return events


def _find_single_trial_start_end_idx(events):
    trial_start_end_markers = [21, 22, 10]
    return np.where(np.isin(events[:, 2], trial_start_end_markers))[0]


def _extract_target_non_target_description(events, run_split_n):
    single_trial_start_end_idx = _find_single_trial_start_end_idx(events)

    n_events = single_trial_start_end_idx.size - 1

    onset_arr = np.empty((n_events,), dtype=np.int64)
    marker_arr = np.empty((n_events,), dtype=np.int64)
    sequence_arr = np.empty((n_events,), dtype=np.int64)
    letter_arr = np.empty((n_events, 12), dtype=np.int64)

    broken_events_idx = list()
    for epoch_idx in range(n_events):
        epoch_start_idx = single_trial_start_end_idx[epoch_idx]
        epoch_end_idx = single_trial_start_end_idx[epoch_idx + 1]

        epoch_events = events[epoch_start_idx:epoch_end_idx]
        onset_ms = _find_epoch_onset(epoch_events)
        if onset_ms == -1:
            broken_events_idx.append(epoch_idx)
            continue

        onset_arr[epoch_idx] = onset_ms
        marker_arr[epoch_idx] = int(
            _single_trial_contains_target(epoch_events)
        )  # 1/true if single trial has target

        sequence_arr[epoch_idx] = _single_trail_sequence_type(epoch_events)
        letter_arr[epoch_idx] = _single_trail_letter_numbers(epoch_events)
    if run_split_n is None or run_split_n == 1:
        return (
            [np.delete(onset_arr, broken_events_idx)],
            [np.delete(marker_arr, broken_events_idx)],
            [np.delete(sequence_arr, broken_events_idx)],
            [np.delete(letter_arr, broken_events_idx, axis=0)],
        )
    else:
        return _split_run_into_n_splits_of_trials(
            broken_events_idx,
            marker_arr,
            onset_arr,
            sequence_arr,
            letter_arr,
            run_split_n,
        )


def _split_run_into_n_splits_of_trials(
    broken_events, marker_arr, onset_arr, sequence_arr, letter_arr, run_split_n
):
    epochs_per_trial = 68
    trials_per_run = 7

    trials_per_split = int(trials_per_run / run_split_n)
    trial_split_idx = (
        np.arange(start=trials_per_split, stop=trials_per_run, step=trials_per_split)
        * epochs_per_trial
    )
    epochs_per_split = trials_per_split * epochs_per_trial

    broken_events = np.array(broken_events)
    has_broken = len(broken_events) > 0

    def broken_idx_for_split(split_idx):
        if not has_broken:
            return []

        broke_int_split_i = broken_events - split_idx * epochs_per_split
        idx_broken_events_in_split = (broke_int_split_i >= 0) & (
            broke_int_split_i < epochs_per_split
        )
        return broke_int_split_i[idx_broken_events_in_split]

    def split_data(data_arr):
        splits = np.split(data_arr, trial_split_idx)
        del splits[run_split_n:]

        return [np.delete(split, broken_idx_for_split(i)) for i, split in enumerate(splits)]

    onset_arr_list = split_data(onset_arr)
    marker_arr_list = split_data(marker_arr)
    sequence_arr_list = split_data(sequence_arr)
    letter_arr_list = split_data(letter_arr)
    return onset_arr_list, marker_arr_list, sequence_arr_list, letter_arr_list


def _find_epoch_onset(epoch_events):
    optical_idx = epoch_events[:, 2] == OPTICAL_MARKER_CODE
    stimulus_onset_time = epoch_events[optical_idx, 0]

    def second_optical_is_feedback():
        if stimulus_onset_time.size != 2:
            return False

        stimulus_prior_second_optical_marker = epoch_events[np.where(optical_idx)[0][1] - 1, 2]
        return stimulus_prior_second_optical_marker in [50, 51, 11]

    if stimulus_onset_time.size == 1 or second_optical_is_feedback():
        return stimulus_onset_time[0]

    # broken epoch: no true onset found..
    return -1


def _single_trial_contains_target(trial_events):
    trial_markers = trial_events[:, 2]
    return np.any((trial_markers > 100) & (trial_markers <= 142))


def _single_trail_sequence_type(trial_events):
    sequence_marker = trial_events[:, 2][0]
    return sequence_marker


def _single_trail_letter_numbers(trial_events):
    sequence_marker = trial_events[:, 2]
    sequence_letter_numbers = []
    for marker in sequence_marker:
        if marker in VALID_LETTER_NUMBERS:
            sequence_letter_numbers.append(marker)

    return np.array(sequence_letter_numbers)



from __future__ import annotations

import time
from typing import TYPE_CHECKING

import numpy as np
import sklearn
import sklearn.linear_model
import sklearn.utils.multiclass
from blockmatrix import (
    SpatioTemporalMatrix,
    fortran_block_levinson,
    fortran_cov_mean_transformation,
    linear_taper,
)
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils.validation import check_is_fitted, check_X_y

from toeplitzlda.classification import ShrinkageLinearDiscriminantAnalysis
from toeplitzlda.classification.covariance import (
    calc_n_times,
    shrinkage,
    subtract_classwise_means,
)


class ExternalLDA(ShrinkageLinearDiscriminantAnalysis):
    """No capabilities except calculating w from external cov/mean information."""

    def __init__(
        self,
        means=None,
        cov=None,
    ):
        self.means = means
        self.cov = cov

    def calc(self):
        if self.means is None or self.cov is None:
            raise ValueError(f"Cannot calculate weights, missing information.")

        w = np.linalg.solve(self.cov, self.means)
        self.classes_ = np.arange(0, w.shape[1])
        # TODO: implement prior_offset
        b = -0.5 * np.sum(self.means * w, axis=0).T  # + prior_offset

        self.coef_ = w.T
        self.intercept_ = b

    def fit(self, X_train, y, oracle_data=None):
        pass



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

def parse_requirements_file(fname):
    requirements = list()
    with open(fname, 'r') as fid:
        for line in fid:
            req = line.strip()
            if req.startswith('#'):
                continue
            # strip end-of-line comments
            req = req.split('#', maxsplit=1)[0].strip()
            requirements.append(req)
    return requirements

#install_requires = parse_requirements_file("requirements.txt")

install_requires=[
    'numpy',
    'sklearn',
    'blockmatrix'
]

setuptools.setup(
    name="pyclf-bci",
    version="0.0.2",
    author="Simon Kojima",
    author_email="simon.kojima@outlook.com",
    description="An implementation of various classifier for bci application.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/simonkojima/pyclf",
    project_urls={
        "Bug Tracker": "https://github.com/simonkojima/pyclf/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires = install_requires,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)


import numpy as np
from blockmatrix import SpatioTemporalMatrix, fortran_block_levinson
from sklearn.base import BaseEstimator

from ..classification.covariance import calc_n_times
from ..classification.toeplitzlda import shrinkage, subtract_classwise_means


class LearningFromLabelProportions(BaseEstimator):
    """
    Learning from label proportions SLDA from Hübner et al 2017 [1]_.

    Parameters
    ----------
    ratio_matrix: np.ndarray of shape (2, 2)
        | ratio seq1/target   ratio seq1/non-target |
        | ratio seq2/target   ratio seq2/non-target |

    References
    ----------
    .. [1] Hübner, D., Verhoeven, T., Schmid, K., Müller, K. R., Tangermann, M., & Kindermans, P. J. (2017)
           Learning from label proportions in brain-computer interfaces: Online unsupervised learning with guarantees.
           PLOS ONE 12(4): e0175856.
           https://doi.org/10.1371/journal.pone.0175856
    """

    def __init__(
        self,
        ratio_matrix: np.ndarray = None,
        toeplitz_time=False,
        taper_time=None,
        toeplitz_spatial=False,
        taper_spatial=None,
        use_fortran_solver=False,
        n_times=None,
        n_channels=31,
    ):
        if ratio_matrix is None:
            self.ratio_matrix = np.array([[3 / 8, 5 / 8], [2 / 18, 16 / 18]])
        self.pinv_ratio_matrix = np.linalg.inv(self.ratio_matrix)
        self.w = None
        self.b = None
        self.n_times = n_times
        self.n_channels = n_channels
        self.toeplitz_time = toeplitz_time
        self.taper_time = taper_time
        self.toeplitz_spatial = toeplitz_spatial
        self.taper_spatial = taper_spatial
        self.use_fortran_solver = use_fortran_solver

        self.mu_T = None
        self.mu_NT = None

        self.stm_info = None

    def fit(self, X, y):
        """
        Parameters
        ----------
        X: np.ndarray
            Input data of shape (n_samples, n_chs, n_time)
        y: np.ndarray
            Sequence labels of X (not target/non-target labels),
             must be 1 (=sequence1) or 2 (=sequence2).
        """
        X1 = X[np.where(y == 1)]
        X2 = X[np.where(y == 2)]

        X = X.reshape(X.shape[0], -1)
        X1 = X1.reshape(X1.shape[0], -1)
        X2 = X2.reshape(X2.shape[0], -1)

        # Compute global covariance matrix
        C_cov, gamma = shrinkage(X.T)
        nt = calc_n_times(C_cov.shape[0], self.n_channels, self.n_times)
        stm = SpatioTemporalMatrix(C_cov, n_chans=self.n_channels, n_times=nt)

        if self.toeplitz_time:
            stm.force_toeplitz_offdiagonals()
        if self.taper_time is not None:
            stm.taper_offdiagonals(self.taper_time)
        stm.swap_primeness()
        if self.toeplitz_spatial:
            stm.force_toeplitz_offdiagonals(raise_spatial=False)
        if self.taper_spatial is not None:
            stm.taper_offdiagonals(self.taper_spatial)
        stm.swap_primeness()

        self.stored_stm = stm
        C_cov = stm.mat

        # Compute sequence-wise means
        average_O1 = np.mean(X1, axis=0)
        average_O2 = np.mean(X2, axis=0)

        mu_T = self.pinv_ratio_matrix[0, 0] * average_O1 + self.pinv_ratio_matrix[0, 1] * average_O2
        mu_NT = (
            self.pinv_ratio_matrix[1, 0] * average_O1 + self.pinv_ratio_matrix[1, 1] * average_O2
        )

        # use reconstructed means to compute w and b
        C_diff = mu_T - mu_NT
        C_mean = 0.5 * (mu_T + mu_NT)

        self.stored_cl_mean = np.vstack([mu_NT, mu_T]).T

        if self.use_fortran_solver:
            if not self.toeplitz_time:
                raise ValueError("Cannot use fortran solver without block-Toeplitz structure")
            C_w = fortran_block_levinson(C_cov, C_diff, nch=self.n_channels, ntim=self.n_times)
        else:
            C_w = np.linalg.solve(C_cov, C_diff)
        C_w = 2 * C_w / np.dot(C_w.T, C_diff)
        C_b = np.dot(-C_w.T, C_mean)

        self.w = C_w
        self.b = C_b

        self.mu_T = mu_T
        self.mu_NT = mu_NT

        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        X = X.reshape(X.shape[0], -1)
        return np.dot(X, self.w) + self.b

    def predict(self, X: np.ndarray):
        return self.decision_function(X) > 0



from typing import Optional, Tuple

import numpy as np
from blockmatrix import SpatioTemporalMatrix, linear_taper
from sklearn import config_context
from sklearn.covariance import LedoitWolf, ledoit_wolf
from sklearn.preprocessing import StandardScaler


def shrinkage(
    X: np.ndarray,
    gamma: Optional[float] = None,
    T: Optional[np.ndarray] = None,
    S: Optional[np.ndarray] = None,
    standardize: bool = True,
) -> Tuple[np.ndarray, float]:
    p, n = X.shape

    if standardize:
        sc = StandardScaler()
        X = sc.fit_transform(X.T).T
    Xn = X - np.repeat(np.mean(X, axis=1, keepdims=True), n, axis=1)
    if S is None:
        S = np.matmul(Xn, Xn.T)
    Xn2 = np.square(Xn)
    idxdiag = np.diag_indices(p)

    nu = np.mean(S[idxdiag])
    if T is None:
        T = nu * np.eye(p, p)

    # Ledoit Wolf
    V = 1.0 / (n - 1) * (np.matmul(Xn2, Xn2.T) - np.square(S) / n)
    if gamma is None:
        gamma = n * np.sum(V) / np.sum(np.square(S - T))
    if gamma > 1:
        print("logger.warning('forcing gamma to 1')")
        gamma = 1
    elif gamma < 0:
        print("logger.warning('forcing gamma to 0')")
        gamma = 0
    Cstar = (gamma * T + (1 - gamma) * S) / (n - 1)
    if standardize:  # scale back
        Cstar = sc.scale_[np.newaxis, :] * Cstar * sc.scale_[:, np.newaxis]

    return Cstar, gamma


def subtract_classwise_means(xTr, y, ext_mean=None):
    n_classes = len(np.unique(y))
    n_features = xTr.shape[0]
    X = np.zeros((n_features, 0))
    cl_mean = np.zeros((n_features, n_classes))
    for ci, cur_class in enumerate(np.unique(y)):
        class_idxs = y == cur_class
        cl_mean[:, ci] = np.mean(xTr[:, class_idxs], axis=1)

        if ext_mean is None:
            X = np.concatenate(
                [
                    X,
                    xTr[:, class_idxs]
                    - np.dot(cl_mean[:, ci].reshape(-1, 1), np.ones((1, np.sum(class_idxs)))),
                ],
                axis=1,
            )
        else:
            X = np.concatenate(
                [
                    X,
                    xTr[:, class_idxs]
                    - np.dot(ext_mean[:, ci].reshape(-1, 1), np.ones((1, np.sum(class_idxs)))),
                ],
                axis=1,
            )
    return X, cl_mean


def calc_n_times(dim, n_channels, n_times):
    if type(n_times) is int:
        return n_times
    elif n_times == "infer":
        if dim % n_channels != 0:
            raise ValueError(f"Could not infer time samples. Remainder is non-zero.")
        else:
            return dim // n_channels
    else:
        raise ValueError(f"Unknown value for n_times: {n_times}")


class ToepTapLW(LedoitWolf):
    """An sklearn compatible covariance estimator"""

    def __init__(
        self,
        n_channels=None,
        *,
        n_times="infer",
        data_is_channel_prime=True,
        standardize=True,
        tapering=linear_taper,
        only_lw=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.standardize = standardize
        self.tapering = tapering
        if self.tapering is None:
            print(
                "WARNING: using block-Toeplitz structure without tapering can lead to numerical "
                "instabilities / singular matrices"
            )
            # Use a no-op taper that always returns 1 as scaling factor
            self.tapering = lambda d, dmax: 1
        self.n_times = n_times
        self.n_channels = n_channels
        self.data_is_channel_prime = data_is_channel_prime
        self.only_lw = only_lw

    def fit(self, X, y=None):
        """Fit the covariance model to X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.
        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        X = self._validate_data(X)
        dim = X.shape[1]
        if self.assume_centered:
            self.location_ = np.zeros(X.shape[1])
        else:
            self.location_ = X.mean(0)
        with config_context(assume_finite=True):
            X_cent = X - self.location_
            sc = None
            if self.standardize:
                sc = StandardScaler()
                X_cent = sc.fit_transform(X_cent)
            covariance, shrinkage_gamma = ledoit_wolf(
                X_cent,
                assume_centered=True,
                block_size=self.block_size,
            )
            # covariance = empirical_covariance(
            #     X_cent, assume_centered=True
            # )
            # covariance, shrinkage = oas(X_cent, assume_centered=True)
            if self.standardize:  # scale back
                covariance = sc.scale_[np.newaxis, :] * covariance * sc.scale_[:, np.newaxis]
        self.shrinkage_ = shrinkage_gamma

        nt = calc_n_times(dim, self.n_channels, self.n_times)
        stm = SpatioTemporalMatrix(
            covariance,
            n_times=nt,
            n_chans=self.n_channels,
            channel_prime=self.data_is_channel_prime,
        )

        if not self.only_lw:
            if not self.data_is_channel_prime:
                stm.swap_primeness()
            stm.force_toeplitz_offdiagonals()
            stm.taper_offdiagonals(self.tapering)
            if not self.data_is_channel_prime:
                stm.swap_primeness()
            covariance = stm.mat
        self.stm_ = stm
        self._set_covariance(covariance)

        return self



from __future__ import annotations

import time
from typing import TYPE_CHECKING

import numpy as np
import sklearn
import sklearn.linear_model
import sklearn.utils.multiclass
from blockmatrix import (
    SpatioTemporalMatrix,
    fortran_block_levinson,
    fortran_cov_mean_transformation,
    linear_taper,
)
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin
from sklearn.utils.extmath import safe_sparse_dot
from sklearn.utils.validation import check_is_fitted, check_X_y

from .covariance import (
    calc_n_times,
    shrinkage,
    subtract_classwise_means,
)

if TYPE_CHECKING:
    from mne import BaseEpochs


class EpochsVectorizer(BaseEstimator, TransformerMixin):
    def __init__(
        self,
        permute_channels_and_time=True,
        select_ival=None,
        jumping_mean_ivals=None,
        averaging_samples=None,
        rescale_to_uv=True,
        mne_scaler=None,
        pool_times=False,
        to_numpy_only=False,
        copy=True,
    ):
        self.permute_channels_and_time = permute_channels_and_time
        self.jumping_mean_ivals = jumping_mean_ivals
        self.select_ival = select_ival
        self.averaging_samples = averaging_samples
        self.rescale_to_uv = rescale_to_uv
        self.scaling = 1e6 if self.rescale_to_uv else 1
        self.pool_times = pool_times
        self.to_numpy_only = to_numpy_only
        self.copy = copy
        self.mne_scaler = mne_scaler

        if self.select_ival is None and self.jumping_mean_ivals is None:
            raise ValueError("jumping_mean_ivals or select_ival is required")

    def fit(self, X, y=None):
        """fit."""
        return self

    def transform(self, X: BaseEpochs):
        """transform."""
        e = X.copy() if self.copy else X
        if self.to_numpy_only:
            X = e.get_data() * self.scaling
            return X
        if self.jumping_mean_ivals is not None:
            self.averaging_samples = np.zeros(len(self.jumping_mean_ivals))
            X = e.get_data() * self.scaling
            new_X = np.zeros((X.shape[0], X.shape[1], len(self.jumping_mean_ivals)))
            for i, ival in enumerate(self.jumping_mean_ivals):
                np_idx = e.time_as_index(ival)
                idx = list(range(np_idx[0], np_idx[1]))
                self.averaging_samples[i] = len(idx)
                new_X[:, :, i] = np.mean(X[:, :, idx], axis=2)
            X = new_X
        elif self.select_ival is not None:
            e.crop(tmin=self.select_ival[0], tmax=self.select_ival[1])
            self.times_ = e.times
            X = e.get_data() * self.scaling
        elif self.pool_times:
            X = e.get_data() * self.scaling
            raise ValueError("This should never be entered though.")
        else:
            raise ValueError("In the constructor, pass either select ival or jumping means.")
        if self.mne_scaler is not None:
            X = self.mne_scaler.fit_transform(X)
        if self.permute_channels_and_time and not self.pool_times:
            X = X.transpose((0, 2, 1))
        if self.pool_times:
            X = np.reshape(X, (-1, X.shape[1]))
        else:
            X = np.reshape(X, (X.shape[0], -1))
        return X


class ShrinkageLinearDiscriminantAnalysis(
    ClassifierMixin,
    BaseEstimator,
):
    """SLDA Implementation with bells and whistles and a lot of options."""

    def __init__(
        self,
        priors=None,
        only_block=False,
        n_times="infer",
        n_channels=31,
        pool_cov=True,
        standardize_shrink=True,
        calculate_oracle_mean=None,
        unit_w=False,
        fixed_gamma=None,
        enforce_toeplitz=False,
        use_fortran_solver=False,
        banding=None,
        tapering=None,
        data_is_channel_prime=True,
    ):
        self.only_block = only_block
        self.priors = priors
        self.n_times = n_times
        self.n_channels = n_channels
        self.pool_cov = pool_cov
        self.standardize_shrink = standardize_shrink
        self.calculate_oracle_mean = calculate_oracle_mean
        self.unit_w = unit_w
        self.fixed_gamma = fixed_gamma
        self.enforce_toeplitz = enforce_toeplitz
        self.use_fortran_solver = use_fortran_solver
        if self.use_fortran_solver and not self.enforce_toeplitz:
            raise ValueError("Can only use Fortran solver when enforce_toeplitz=True")
        self.banding = banding
        self.tapering = tapering
        self.data_is_channel_prime = data_is_channel_prime

        # added for adoptation
        self.C_inv = None
        self.cl_mean = None
        self.y_train = None
        self.prior_offset = None

    def fit(self, X_train, y, oracle_data=None):
        # Section: Basic setup
        check_X_y(X_train, y)
        self.classes_ = sklearn.utils.multiclass.unique_labels(y)
        if self.calculate_oracle_mean is None:
            oracle_data = None
        xTr = X_train.T

        self.y_train = y

        if self.priors is None:
            # here we deviate from the bbci implementation and
            # use the sample priors by default
            _, y_t = np.unique(y, return_inverse=True)  # non-negative ints
            priors = np.bincount(y_t) / float(len(y))
        else:
            priors = self.priors

        # Section: covariance / mean calculation
        X, cl_mean = subtract_classwise_means(xTr, y)
        # Check if mean information is provided and if so whether to use it only for class-wise
        # mean estimation or also for the subtraction of class means from the data to estimate the
        # covariance matrix
        if self.calculate_oracle_mean == "clmean_and_covmean":
            _, cl_mean = subtract_classwise_means(oracle_data["x"].T, oracle_data["y"])
            X, _ = subtract_classwise_means(xTr, y, ext_mean=cl_mean)
        elif self.calculate_oracle_mean == "only_clmean":
            _, cl_mean = subtract_classwise_means(oracle_data["x"].T, oracle_data["y"])
        # Subtract class-wise means from data and then estimate the covariance or average
        # class-wise covariance matrices?
        if self.pool_cov:
            C_cov, C_gamma = shrinkage(
                X,
                standardize=self.standardize_shrink,
                gamma=self.fixed_gamma,
            )
        else:
            C_cov = np.zeros((xTr.shape[0], xTr.shape[0]))
            for cur_class in self.classes_:
                class_idxs = y == cur_class
                x_slice = X[:, class_idxs]
                C_cov += priors[cur_class] * shrinkage(x_slice)[0]
        dim = C_cov.shape[0]
        nt = calc_n_times(dim, self.n_channels, self.n_times)
        stm = SpatioTemporalMatrix(
            C_cov,
            n_chans=self.n_channels,
            n_times=nt,
            channel_prime=self.data_is_channel_prime,
        )
        if not self.data_is_channel_prime:
            stm.swap_primeness()
        if self.enforce_toeplitz:
            stm.force_toeplitz_offdiagonals()
        # Banding could be realized with a binary taper, so remove?
        if self.banding is not None:
            stm.band_offdiagonals(self.banding)
        if self.tapering is not None:
            stm.taper_offdiagonals(taper_f=self.tapering)
        if not self.data_is_channel_prime:
            stm.swap_primeness()

        self.stored_cl_mean = cl_mean
        self.stored_stm = stm
        C_cov = stm.mat

        if self.only_block:
            C_cov_new = np.zeros_like(C_cov)
            for i in range(nt):
                idx_start = i * self.n_channels
                idx_end = idx_start + self.n_channels
                C_cov_new[idx_start:idx_end, idx_start:idx_end] = C_cov[
                    idx_start:idx_end, idx_start:idx_end
                ]
            C_cov = C_cov_new

        prior_offset = np.log(priors)

        # Numpy uses more cores, fortran library only one
        # with threadpoolctl.threadpool_limits(limits=1):
        if self.use_fortran_solver:
            # TODO implement transformation for mean.ndim == 2 in block_matrix.py
            w = np.empty_like(cl_mean)
            self.fit_time_ = 0
            for ci in range(w.shape[1]):
                fcov, fmean = fortran_cov_mean_transformation(
                    C_cov, cl_mean[:, ci], nch=self.n_channels, ntim=nt
                )
                st = time.time()
                w[:, ci] = fortran_block_levinson(fcov, fmean, transform_A=False)
                self.fit_time_ += time.time() - st
        else:
            st = time.time()
            w = np.linalg.solve(C_cov, cl_mean)
            self.fit_time_ = time.time() - st
        
        w = w / np.linalg.norm(w) if self.unit_w else w
        b = -0.5 * np.sum(cl_mean * w, axis=0).T + prior_offset

        self.coef_ = w.T
        self.intercept_ = b
        
        self.C_inv = np.linalg.inv(C_cov)
        self.cl_mean = cl_mean

        return C_cov, cl_mean, prior_offset
    
    def adaptation(self, X, y, eta_cov = 0.001, eta_mean = 0.005):
        """
        Calculate new w and b by adaptation method proposed by Vidaurre et al.(2011)[1]

        Parameters
        ----------
        X : {array-like} of shape (n_samples, n_features)
            The feature matrix to be feeded for adaptation.


        [1] Vidaurre et al. Toward Unsupervised Adaptation of LDA for Brain_Computer Interfaces. (2011)
        
        """
        
        y_train = np.append(self.y_train, y)
        C_inv = self.C_inv
        cl_mean = self.cl_mean

        for idx, cl in enumerate(y):
            
            x = X[idx, :]
            x = x.reshape((x.size, 1))

            # Covariance adaptation
            v = np.dot(C_inv, x)
            
            frac_num = np.dot(v, v.T)
            frac_den = ((1 - eta_cov)/eta_cov) + np.dot(x.T, v)
            
            I = C_inv - (frac_num / frac_den)
            C_inv = I / (1 - eta_cov)

            # Class mean adaptation
            class_idx = np.where(self.classes_ == cl)[0]
            
            shape = cl_mean[:, class_idx].shape

            cl_mean[:,class_idx] = (1-eta_mean)*cl_mean[:,class_idx] + (eta_mean*x)
         
            
        if self.priors is None:
            # here we deviate from the bbci implementation and
            # use the sample priors by default
            _, y_t = np.unique(y, return_inverse=True)  # non-negative ints
            priors = np.bincount(y_t) / float(len(y))
        else:
            priors = self.priors
        prior_offset = np.log(priors)

        self.C_inv = C_inv
        self.cl_mean = cl_mean
        self.y_train
        self.prior_offset = prior_offset
    
    def apply_adaptation(self):

        w = np.dot(self.C_inv, self.cl_mean)
        w = w / np.linalg.norm(w) if self.unit_w else w
        b = -0.5 * np.sum(self.cl_mean * w, axis=0).T + self.prior_offset

        self.coef_ = w.T
        self.intercept_ = b

    def decision_function(self, X):
        """
        Predict confidence scores for samples.

        The confidence score for a sample is proportional to the signed
        distance of that sample to the hyperplane.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the confidence scores.

        Returns
        -------
        scores : ndarray of shape (n_samples,) or (n_samples, n_classes)
            Confidence scores per `(n_samples, n_classes)` combination. In the
            binary case, confidence score for `self.classes_[1]` where >0 means
            this class would be predicted.
        """
        check_is_fitted(self)

        X = self._validate_data(X, accept_sparse="csr", reset=False)
        scores = safe_sparse_dot(X, self.coef_.T, dense_output=True) + self.intercept_
        if scores.shape[1] == 2:
            scores = scores[:, 1] - scores[:, 0]
        return scores.squeeze()

    def predict(self, X):
        """
        Predict class labels for samples in X.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The data matrix for which we want to get the predictions.

        Returns
        -------
        y_pred : ndarray of shape (n_samples,)
            Vector containing the class labels for each sample.
        """
        scores = self.decision_function(X)
        if len(scores.shape) == 1:
            indices = (scores > 0).astype(int)
        else:
            indices = scores.argmax(axis=1)
        return self.classes_[indices]

    def predict_proba(self, X):
        """Estimate probability.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data.

        Returns
        -------
        C : array, shape (n_samples, n_classes)
            Estimated probabilities.
        """
        prob = self.decision_function(X)
        prob *= -1
        np.exp(prob, prob)
        prob += 1
        np.reciprocal(prob, prob)

        return np.column_stack([1 - prob, prob])

    def predict_log_proba(self, X):
        """Estimate log probability.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Input data.

        Returns
        -------
        C : array, shape (n_samples, n_classes)
            Estimated log probabilities.
        """
        return np.log(self.predict_proba(X))


class ToeplitzLDA(ShrinkageLinearDiscriminantAnalysis):
    def __init__(
        self,
        n_channels=None,
        *,
        n_times="infer",
        data_is_channel_prime=True,
        use_fortran_solver=False,
    ):
        if n_channels is None:
            raise ValueError(f"Required parameter n_channels is not set.")
        super().__init__(
            data_is_channel_prime=data_is_channel_prime,
            n_times=n_times,
            n_channels=n_channels,
            use_fortran_solver=use_fortran_solver,
            tapering=linear_taper,
            enforce_toeplitz=True,
        )


# This is a very simple implementation of sLDA for didactic purposes
class PlainLDA(BaseEstimator, ClassifierMixin):
    """Straightforward SLDA implementation mostly for ease of understanding.

    In contrast to the other LDAs in this file, this implementation actually uses the bias for
    class prediction. The other classifiers report the class with maximal classifier output, i.e.,
    output from decision_function."""

    def __init__(
        self,
        toeplitz_time=False,
        taper_time=None,
        use_fortran_solver=False,
        n_times=None,
        n_channels=None,
        global_cov=False,
    ):
        self.w = None
        self.b = None
        self.n_times = n_times
        self.n_channels = n_channels
        self.toeplitz_time = toeplitz_time
        self.taper_time = taper_time
        self.use_fortran_solver = use_fortran_solver
        self.global_cov = global_cov

        self.mu_T = None
        self.mu_NT = None

        self.stm_info = None

    def fit(self, X, y):
        """
        Parameters
        ----------
        X: np.ndarray
            Input data of shape (n_samples, n_chs, n_time)
        y: np.ndarray
            Actual labels of X
        """

        X = X.reshape(X.shape[0], -1)
        self.classes_ = np.unique(y)
        if set(self.classes_) != {0, 1}:
            raise ValueError(
                "This LDA only supports binary classification. Use one of the other classifiers "
                "for multi-class settings."
            )

        mu_T = np.mean(X[np.where(y == 1)], axis=0)
        mu_NT = np.mean(X[np.where(y == 0)], axis=0)

        # Subtract class-means unless we want to ignore them
        if not self.global_cov:
            X = subtract_classwise_means(X.T, y)[0].T
        C_cov, gamma = shrinkage(X.T)

        # ToeplitzLDA specific code BEGIN
        nt = calc_n_times(C_cov.shape[0], self.n_channels, self.n_times)
        stm = SpatioTemporalMatrix(C_cov, n_chans=nt, n_times=self.n_times)

        if self.toeplitz_time:
            stm.force_toeplitz_offdiagonals()
        if self.taper_time is not None:
            stm.taper_offdiagonals(self.taper_time)

        self.stored_stm = stm
        C_cov = stm.mat
        # ToeplitzLDA specific code END

        C_diff = mu_T - mu_NT
        C_mean = 0.5 * (mu_T + mu_NT)

        if self.use_fortran_solver:
            if not self.toeplitz_time:
                raise ValueError("Cannot use fortran solver without block-Toeplitz structure")
            C_w = fortran_block_levinson(C_cov, C_diff, nch=self.n_channels, ntim=self.n_times)
        else:
            C_w = np.linalg.solve(C_cov, C_diff)
        C_b = np.dot(-C_w.T, C_mean)

        self.w = C_w.reshape((1, -1))
        self.b = C_b

        self.mu_T = mu_T
        self.mu_NT = mu_NT

        return self

    def decision_function(self, X: np.ndarray) -> np.ndarray:
        X = X.reshape(X.shape[0], -1)
        return np.dot(X, self.w.T) + self.b

    def predict(self, X: np.ndarray):
        return self.decision_function(X) > 0
